const { expect } = require('chai')
const sinon = require('sinon')
const Container = require('../src/Container')
const util = require('util')

describe('the inversion of control container', function() {
    it('should be a functional class', function() {
        expect(Container).to.be.a('function')
    })

    it('should be constructable', function() {
        expect(new Container).to.be.an.instanceof(Container)
    })

    it('should accept definitions as parameters on construction', function() {
        const example_defs = {}
        const container = new Container(example_defs)

        expect(container.definitions).to.be.equal(example_defs)
    })

    it('should set an empty object of definitions by default', function() {
        const container = new Container()
        expect(container.definitions).to.be.an('object')
        expect(Object.keys(container.definitions).length).to.be.equal(0)
    })

    it('should have a deferred_classes array', function() {
        const container = new Container()
        expect(container.deferred_classes).to.be.an('array')
    })

    it('should have a statics object', function() {
        const container = new Container()
        expect(container.statics).to.be.an('object')
    })

    it('should have an instances object', function() {
        const container = new Container()
        expect(container.instances).to.be.an('object')
    })

    describe('has function', function() {
        it('should be a function', function() {
            const container = new Container()
            expect(container.has).to.be.a('function')
        })

        it('should accept 1 argument', function() {
            const container = new Container()
            expect(container.has.length).to.be.equal(1)
        })

        it('should return true if a service definition exists with that name', function() {
            const name = Math.random().toString(36).substring(7)
            const defs = {}
            defs[name] = {}
            const container = new Container(defs)
            expect(container.has(name)).to.be.equal(true)
        })

        it('should return false if a service definition does not exist with that name', function() {
            const name = Math.random().toString(36).substring(7)
            const container = new Container({})
            expect(container.has(name)).to.be.equal(false)
        })
    })

    describe('service function', function() {
        it('should be a function', function() {
            const container = new Container()
            expect(container.service).to.be.a('function')
        })

        it('should require no arguments', function() {
            const container = new Container()
            expect(container.service.length).to.be.equal(0)
        })

        it('should throw an error with a helpful message if the passed in service has no definition', function() {
            const container = new Container()
            const name = Math.random().toString(36).substring(7)
            expect(container.service.bind(container, name)).to.throw('No such service registered with this container: '+name)
        })

        it('should instantiate service classes by name', function() {
            class TestService {}
            const name = Math.random().toString(36).substring(7)
            const defs = {}
            defs[name] = TestService
            const container = new Container(defs)
            expect(container.service(name)).to.be.an.instanceof(TestService)
        })

        it('should re-use instantiated services', function() {
            class TestService {}
            const name = Math.random().toString(36).substring(7)
            const defs = {}
            defs[name] = TestService
            const container = new Container(defs)
            const test_service_1 = container.service(name)
            const test_service_2 = container.service(name)
            expect(test_service_1).to.be.equal(test_service_2)
        })

        it('should inject services before instantiation', function() {
            class TestService {}
            const fake_di = { make: sinon.spy() }
            const defs = { test: TestService }
            const container = new Container(defs)
            container.di = fake_di
            container.service('test')

            expect(fake_di.make.calledOnce).to.be.true
            expect(fake_di.make.getCall(0).args.length).to.be.equal(1)
            expect(fake_di.make.getCall(0).args[0]).to.be.equal(TestService)
        })

        it('should return a proxy container if called with no arguments', function() {
            class TestService {}
            const name = Math.random().toString(36).substring(7)
            const defs = {}
            defs[name] = TestService
            const container = new Container(defs)
            const services = container.service()

            expect(util.types.isProxy(services)).to.be.true
            expect(services[name]).to.be.an.instanceof(TestService)
        })
    })

    describe('proxy function', function() {
        it('should return a proxy container that can resolve services', function() {
            class TestService {}
            const name = Math.random().toString(36).substring(7)
            const defs = {}
            defs[name] = TestService
            const container = new Container(defs)
            const services = container.proxy()

            expect(util.types.isProxy(services)).to.be.true
            expect(services[name]).to.be.an.instanceof(TestService)
        })
    })

    describe('register function', function() {
        it('should be a function', function() {
            const container = new Container()
            expect(container.register).to.be.a('function')
        })

        it('should accept 2 arguments', function() {
            const container = new Container()
            expect(container.register.length).to.be.equal(2)
        })

        it('should register service definitions by name', function() {
            class TestService {}
            const name = Math.random().toString(36).substring(7)
            const container = new Container()

            container.register(name, TestService)
            expect(container.definitions[name]).to.be.equal(TestService)
        })

        it('should trigger deferral processing', function() {
            class TestService {}
            const name = Math.random().toString(36).substring(7)
            const container = new Container()

            container._process_deferral = sinon.spy()
            container.register(name, TestService)
            expect(container._process_deferral.calledOnce).to.be.true
            expect(container._process_deferral.getCall(0).args.length).to.be.equal(2)
            expect(container._process_deferral.getCall(0).args[0]).to.be.equal(name)
            expect(container._process_deferral.getCall(0).args[1]).to.be.an.instanceof(TestService)
        })
    })

    describe('register_as_instance function', function() {
        it('should be a function', function() {
            const container = new Container()
            expect(container.register_as_instance).to.be.a('function')
        })

        it('should accept 2 arguments', function() {
            const container = new Container()
            expect(container.register_as_instance.length).to.be.equal(2)
        })

        it('should register service constructors as definitions by name', function() {
            class TestService {}
            const name = Math.random().toString(36).substring(7)
            const container = new Container()
            const test = new TestService()

            container.register_as_instance(name, test)
            expect(container.definitions[name]).to.be.equal(TestService)
        })

        it('should register service instances by name for reuse', function() {
            class TestService {}
            const name = Math.random().toString(36).substring(7)
            const container = new Container()
            const test = new TestService()

            container.register_as_instance(name, test)
            expect(container.service(name)).to.be.equal(test)
            expect(container.service(name)).to.be.equal(test)
        })

        it('should trigger deferral processing', function() {
            class TestService {}
            const name = Math.random().toString(36).substring(7)
            const container = new Container()
            const test = new TestService()

            container._process_deferral = sinon.spy()
            container.register_as_instance(name, test)
            expect(container._process_deferral.calledOnce).to.be.true
            expect(container._process_deferral.getCall(0).args.length).to.be.equal(2)
            expect(container._process_deferral.getCall(0).args[0]).to.be.equal(name)
            expect(container._process_deferral.getCall(0).args[1]).to.be.equal(test)
        })
    })

    describe('_process_deferral function', function() {
        let container
        let SomeClass
        let service_name
        let ServiceClass
        beforeEach(function() {
            container = new Container()
            class TestClass {
                static __has_deferred_services = false
                static _di_deferred_services = []
                static __deferral_callback = sinon.spy()
            }
            SomeClass = TestClass
            container.deferred_classes = [TestClass]

            class Service {}
            ServiceClass = Service

            service_name = Math.random().toString(36).substring(7)
        })

        it('should be a function', function() {
            expect(container._process_deferral).to.be.a('function')
        })

        it('should accept 2 arguments', function() {
            expect(container._process_deferral.length).to.be.equal(2)
        })

        it('should remove classes from deferred_classes if they no longer have deferred services', function() {
            container._process_deferral(service_name, ServiceClass)
            expect(container.deferred_classes.length).to.be.equal(0)
        })

        it('should call the __deferral_callback if _di_deferred_services has the service name', function() {
            SomeClass._di_deferred_services.push(service_name)
            container._process_deferral(service_name, ServiceClass)
            expect(SomeClass.__deferral_callback.calledOnce).to.be.true
            expect(SomeClass.__deferral_callback.getCall(0).args.length).to.be.equal(2)
            expect(SomeClass.__deferral_callback.getCall(0).args[0]).to.be.equal(service_name)
            expect(SomeClass.__deferral_callback.getCall(0).args[1]).to.be.equal(ServiceClass)
        })

        it('should not call the __deferral_callback otherwise', function() {
            container._process_deferral(service_name, ServiceClass)
            expect(SomeClass.__deferral_callback.calledOnce).to.be.false
        })

        it('should leave the class in deferred_services if the class still __has_deferred_services', function() {
            SomeClass.__has_deferred_services = true
            container._process_deferral(service_name, ServiceClass)
            expect(container.deferred_classes).to.include(SomeClass)
        })
    })

    describe('defer function', function() {
        let container
        beforeEach(function() {
            container = new Container()
        })

        it('should be a function', function() {
            expect(container.defer).to.be.a('function')
        })

        it('should accept 1 argument', function() {
            expect(container.defer.length).to.be.equal(1)
        })

        it('should check if the class is deferrable', function() {
            class TestClass {}
            const fake = sinon.fake.returns(true)
            container.__is_deferrable = fake
            container.defer(TestClass)

            expect(fake.callCount).to.be.at.least(1)
            expect(fake.lastArg).to.be.equal(TestClass)
        })

        it('should fail if the class is not deferrable', function() {
            class TestClass {}
            const fake = sinon.fake.returns(false)
            container.__is_deferrable = fake

            expect(container.defer.bind(container, TestClass)).to.throw('Cannot defer non-deferrable class: '+TestClass.name)
        })

        it('should add the class to deferred_classes if the class is deferrable', function() {
            class TestClass {}
            const fake = sinon.fake.returns(true)
            container.__is_deferrable = fake
            container.defer(TestClass)

            expect(container.deferred_classes.length).to.be.equal(1)
            expect(container.deferred_classes).to.include(TestClass)
        })
    })

    describe('__is_deferrable function', function() {
        let container
        let fake_class

        beforeEach(function() {
            container = new Container()
            fake_class = {
                _di_deferred_services: [],
                _di_deferred_instances: [],
                _di_allow_defer: true,
                __deferral_callback: () => {},
                __has_deferred_services: false,
            }
        })

        it('should be a function', function() {
            expect(container.__is_deferrable).to.be.a('function')
        })

        it('should accept 1 argument', function() {
            expect(container.__is_deferrable.length).to.be.equal(1)
        })

        it('should require the class to have _di_deferred_services array', function() {
            expect(container.__is_deferrable(fake_class)).to.be.true
            fake_class._di_deferred_services = true
            expect(container.__is_deferrable(fake_class)).to.be.false
        })

        it('should require the class to have _di_deferred_instances array', function() {
            expect(container.__is_deferrable(fake_class)).to.be.true
            fake_class._di_deferred_instances = true
            expect(container.__is_deferrable(fake_class)).to.be.false
        })

        it('should make sure the _di_allow_defer field exists', function() {
            expect(container.__is_deferrable(fake_class)).to.be.true
            delete fake_class._di_allow_defer
            expect(container.__is_deferrable(fake_class)).to.be.false
        })

        it('should verify that the __deferral_callback field is a function', function() {
            expect(container.__is_deferrable(fake_class)).to.be.true
            fake_class.__deferral_callback = true
            expect(container.__is_deferrable(fake_class)).to.be.false
        })

        it('should make sure that the __has_deferred_services field exists', function() {
            expect(container.__is_deferrable(fake_class)).to.be.true
            delete fake_class.__has_deferred_services
            expect(container.__is_deferrable(fake_class)).to.be.false
        })
    })
})
